#include "ultra-cp.h"

//====================
//      PRIVATE
//====================
/**
 * @brief Get the entry permissions as int
 *
 * @param path Entry path
 *
 * @returns Permissions
 */
int get_int_permissions(const char *path) {
    struct stat entry_stat;

    if (stat(path, &entry_stat) < 0)
        return -1;

    return entry_stat.st_mode;
}

/**
 * @brief Copy a single file
 *
 * @param source Source file
 * @param destination Destination file
 */
void copy_single_file(const char *source, const char *destination) {
    int dst_path_lenght = 0;
    char dst_path[PATH_MAX];

    // Check if the basename already exists in the destination path
    if (strstr(destination, basename((char *)source)) == NULL) {
        dst_path_lenght = snprintf(dst_path, PATH_MAX, "%s/%s", destination,
                                   basename((char *)source));
    } else {
        dst_path_lenght = strlen(destination);
        strcpy(dst_path, destination);
    }

    // Path length exceeded clause
    if (dst_path_lenght >= PATH_MAX) {
        fprintf(stderr, "Path length has got too long.\n");
        exit(EXIT_FAILURE);
    }

    FILE *src_file, *dst_file;
    src_file = fopen(source, "rb");
    dst_file = fopen(dst_path, "wb");

    size_t n;
    char buff[BUFSIZ];
    while ((n = fread(buff, 1, BUFSIZ, src_file)) != 0)
        fwrite(buff, 1, n, dst_file);

    fclose(src_file);
    fclose(dst_file);

    chmod(destination, get_int_permissions(source));
}

/**
 * @brief Copy a directory
 *
 * @param source Source directory
 * @param destination Destination directory
 */
void copy_directory(const char *source, const char *destination) {
    mkdir(destination, get_int_permissions(source));
}

/**
 * @brief Copy a source entry to a destination location
 *
 * @param source Source entry
 * @param destination Destination location
 */
void copy_entry(struct dirent *entry, const char *source,
                const char *destination) {
    struct stat entry_stat;

    if (stat(source, &entry_stat) < 0)
        return;

    // Entry is a file, copy it
    if (entry->d_type == DT_REG)
        copy_single_file(source, destination);

    // Go deeper if directory
    if (entry->d_type == DT_DIR)
        copy_directory(source, destination);

    struct utimbuf new_time;
    new_time.actime = entry_stat.st_atim.tv_sec;
    new_time.modtime = entry_stat.st_mtim.tv_sec;

    utime(destination, &new_time);
}

/**
 * @brief Copy the content of a directory to a destination
 *
 * @param source Source directory
 * @param destination Destination directory
 */
void copy_directory_content(const char *source, const char *destination) {
    DIR *dir;
    struct dirent *entry;

    // Cannot open dir
    if (!(dir = opendir(source)))
        return;

    // Loop through all entries of the given directory
    while ((entry = readdir(dir)) != NULL) {
        const char *d_name;
        d_name = entry->d_name;

        // Skip "." and ".."
        if (strcmp(d_name, ".") == 0 || strcmp(d_name, "..") == 0)
            continue;

        int dst_path_lenght;
        char dst_path[PATH_MAX];
        char src_path[PATH_MAX];

        // Concatenate entry name to source and destination directory
        snprintf(src_path, PATH_MAX, "%s/%s", source, d_name);
        dst_path_lenght =
            snprintf(dst_path, PATH_MAX, "%s/%s", destination, d_name);

        // Path length exceeded clause
        if (dst_path_lenght >= PATH_MAX) {
            fprintf(stderr, "Path length has got too long.\n");
            exit(EXIT_FAILURE);
        }

        copy_entry(entry, src_path, dst_path);

        // Go deeper
        if (entry->d_type == DT_DIR)
            copy_directory_content(src_path, dst_path);
    }

    closedir(dir);
}

//====================
//      PUBLIC
//====================
char *get_permissions(const char *name) {
    int perms = get_int_permissions(name);
    char *permissions = calloc(10, sizeof(char));

    permissions[0] = S_ISDIR(perms) ? 'd' : S_ISLNK(perms) ? 'l' : '-';
    permissions[1] = (perms & S_IRUSR) ? 'r' : '-';
    permissions[2] = (perms & S_IWUSR) ? 'w' : '-';
    permissions[3] = (perms & S_IXUSR) ? 'x' : '-';
    permissions[4] = (perms & S_IRGRP) ? 'r' : '-';
    permissions[5] = (perms & S_IWGRP) ? 'w' : '-';
    permissions[6] = (perms & S_IXGRP) ? 'x' : '-';
    permissions[7] = (perms & S_IROTH) ? 'r' : '-';
    permissions[8] = (perms & S_IWOTH) ? 'w' : '-';
    permissions[9] = (perms & S_IXOTH) ? 'x' : '-';

    return permissions;
}

int get_size(const char *name) {
    int size = 0;
    struct stat file_stat;

    // Read stats
    if (stat(name, &file_stat) < 0)
        return 0;

    size = file_stat.st_size;

    return size;
}

char *get_date(const char *name) {
    struct stat file_stat;

    // Read stats
    if (stat(name, &file_stat) < 0)
        return NULL;

    char *date = calloc(25, sizeof(char));

    strftime(date, 25, "%a %b %d %X %Y", localtime(&(file_stat.st_ctime)));

    return date;
}

void list_dir(const char *name) {
    DIR *dir;
    struct dirent *entry;

    // Cannot open dir
    if (!(dir = opendir(name)))
        return;

    // Loop through all entries of the given directory
    while ((entry = readdir(dir)) != NULL) {
        const char *d_name;
        d_name = entry->d_name;

        // Skip "." and ".."
        if (strcmp(d_name, ".") == 0 || strcmp(d_name, "..") == 0)
            continue;

        int path_lenght;
        char path[PATH_MAX];

        // Create the full relative path and file stats
        path_lenght = snprintf(path, PATH_MAX, "%s/%s", name, d_name);
        char *permissions = get_permissions(path);
        int size = get_size(path);
        char *date = get_date(path);

        printf("%s %*d %s  %s\n", permissions, 10, size, date, path);

        free(permissions);
        free(date);

        // Path length exceeded clause
        if (path_lenght >= PATH_MAX) {
            fprintf(stderr, "Path length has got too long.\n");
            exit(EXIT_FAILURE);
        }

        // Go deeper if directory
        if (entry->d_type == DT_DIR)
            list_dir(path);
    }

    closedir(dir);
}

void copy_files(const char **sources, const char *destination, uint8_t options,
                int source_count) {
    for (int s = 0; s < source_count; s++) {
        struct stat source_stat;

        if (stat(sources[s], &source_stat) < 0)
            continue;

        if (S_ISDIR(source_stat.st_mode)) {
            int dst_path_lenght;
            char dst_path[PATH_MAX];

            // Create the full source and destination paths
            dst_path_lenght =
                snprintf(dst_path, PATH_MAX, "%s/%s", destination, basename((char *)sources[s]));

            // Path length exceeded clause
            if (dst_path_lenght >= PATH_MAX) {
                fprintf(stderr, "Path length has got too long.\n");
                exit(EXIT_FAILURE);
            }

            copy_directory(sources[s], dst_path);
            copy_directory_content(sources[s], dst_path);
        }

        if (S_ISREG(source_stat.st_mode))
            copy_single_file(sources[s], destination);
    }

    list_dir(destination);
}
